/*  Copyright (C) 2023 Tibor 'Igor2' Palinkas - license: Creative Commons CC0 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#include <libgpcogl/libgpcogl.h>
#include <libgpcogl/gli.h>

/* Input matrix to scale */
#define SX 4
#define SY 4
const float input_mx[SX*SY] = {
	1,2,1,0,
	3,3,1,5,
	7,3,4,2,
	2,5,6,2
};


#define NL "\n"

const char *mx_scale =
	NL GPCOGL_SHADER_PROG(130)                /* program header, require glsl ver 1.3 */
	NL GPCOGL_OUT_1_FLOAT(output_cell)        /* specify output cell type (1*float), create a variable output_cell for the output */
	NL GPCOGL_IN_1_FLOAT(in_mx)               /* specify an input array with cells (1*float); this is where the current state is read from */
	NL GPCOGL_CONST_1_FLOAT(scl)              /* declare a global const; this isthe scale factor */
	NL "void main()"                          /* shader program entry point */
	NL "{"
	NL "	" GPCOGL_OUT_COORDS(x, y)           /* fetch output cell coords into variables x and y */
	NL "	float inp = getxy_in_mx(x, y);"     /* read in_mx cell in a temp var so the computation is easier to read */
	NL "	output_cell = inp * scl;"           /* the actual calculation */
	NL "}";

static void print_mx(float *f, int sx, int sy)
{
	int x, y;

	for(y = 0; y < sy; y++) {
		for(x = 0; x < sx; x++) {
			if (x > 0)
				printf(" ");
			printf("%.4f", *f);
			f++;
		}
		printf("\n");
	}
}

int main(int argc, char **argv)
{
	gpcogl_prog_id_t prg; /* handle of the shader program */
	gpcogl_t ctx;         /* lib context all arrays and shader programs are in */
	gpcogl_array_id_t ot; /* handle of the output array (GPU-side) */
	gpcogl_array_id_t ia; /* handle of the input array (GPU-side) */
	float *res;           /* CPU-side copy of the output array */
	float mult = 1.3;     /* multiplier to use for scaling matrix elements */

	/* Initialize the opengl interface library (e.g. glut or noxgl) */
	gpcogl_gli_init(&argc, argv);

	/* Create a context (hidden X window with glut) for off-screen rendering */
	if (gpcogl_gli_make_context(&ctx, NULL) != 0)
		return 1;

	/* Compile the shader program from mx_scale, remember it as prg */
	prg = gpcogl_compile(&ctx, mx_scale);
	if (prg == 0) {
		fprintf(stderr, "Failed to compile mx_scale\n");
		return 1;
	}

	/* upload the input array */
	ia = gpcogl_array_init(&ctx, GPCOGL_CT_1_FLOAT32, input_mx, SX, SY);

	/* create a GPU array (would be the output of compute) */
	ot = gpcogl_array_create(&ctx, GPCOGL_CT_1_FLOAT32, SX, SY);

	/* Compute: use precompiled program "prg", output to "ot", no input arrays,
	   constant 2*int "origin" set to org */
	{
		const char *src_name[2];
		gpcogl_array_id_t src_aid[2];
		const char *const_type[2];
		const char *const_name[2];
		const void *const_val[2];

		src_name[0] = "in_mx";
		src_name[1] = NULL;
		src_aid[0] = ia;

		const_type[0] = GPCOGL_C_1F;
		const_type[1] = NULL;
		const_name[0] = "scl";
		const_val[0] = &mult;

		gpcogl_acompute(&ctx, prg, ot, src_name, src_aid, const_type, const_name, const_val);
	}

	/* Allocate memory in CPU ram for the result and download from "ot" to res */
	res = calloc(SX*SY, sizeof(float));
	gpcogl_array_get(&ctx, res, ot);

	/* print result */
	print_mx(res, SX, SY);

	/* free context and lib globals */
	gpcogl_gli_destroy_context(&ctx);
	gpcogl_gli_uninit();

	return 0;
}
