/*
    libgpcogl - Generic Purpose Computation on OpenGL (using shaders)
    Copyright (C) 2023 Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://www.repo.hu/projects/libgpcogl
      Lead developer: http://www.repo.hu/projects/libgpcogl/contact.html
*/


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <GL/glew.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>

#include "libgpcogl.h"

#define NL "\n"
#define false 0

const int gpcogl_ver_major = VER_MAJOR;
const int gpcogl_ver_minor = VER_MINOR;
const int gpcogl_ver_patch = VER_PATCH;

const gpcogl_array_id_t gpcogl_no_array = {-1, 0};

/* Output is always bound to texture unit 0 */
#define TEXTUNIT_STDOUT 0

static gpcogl_t *curr = NULL;

#define MAKE_CURRENT \
	do { \
		if (curr != ctx) { \
			ctx->make_current(ctx); \
			curr = ctx; \
		} \
	} while(0)

#define chkerr(callname, error_ret_val) \
	do { \
		int err = glGetError(); \
		if (err != 0) { \
			fprintf(stderr, "%s: opengl error %x\n", callname, err); \
			return (error_ret_val); \
		} \
	} while(0)


typedef struct gpcogl_cell_spec_s {
	GLenum intfmt;   /* sized internal format */
	GLenum txtfmt;   /* texture format format */
	GLenum type;     /* pixel data type */
} gpcogl_cell_spec_t;

static const gpcogl_cell_spec_t cell_spec[] = {
	/* GPCOGL_CT_1_FLOAT32 */ { GL_R32F, GL_RED, GL_FLOAT },
	/* GPCOGL_CT_2_FLOAT32 */ { GL_RG32F, GL_RG, GL_FLOAT },
	/* GPCOGL_CT_3_FLOAT32 */ { GL_RGB32F, GL_RGB, GL_FLOAT },
	/* GPCOGL_CT_4_FLOAT32 */ { GL_RGBA32F, GL_RGBA, GL_FLOAT },
	/* GPCOGL_CT_1_INT32 */   { GL_R32I, GL_RED_INTEGER, GL_INT },
	/* GPCOGL_CT_2_INT32 */   { GL_RG32I, GL_RG_INTEGER, GL_INT },
	/* GPCOGL_CT_3_INT32 */   { GL_RGB32I, GL_RGB_INTEGER, GL_INT },
	/* GPCOGL_CT_4_INT32 */   { GL_RGBA32I, GL_RGBA_INTEGER, GL_INT },
	/* GPCOGL_CT_1_UINT32 */  { GL_R32UI, GL_RED_INTEGER, GL_UNSIGNED_INT },
	/* GPCOGL_CT_2_UINT32 */  { GL_RG32UI, GL_RG_INTEGER, GL_UNSIGNED_INT },
	/* GPCOGL_CT_3_UINT32 */  { GL_RGB32UI, GL_RGB_INTEGER, GL_UNSIGNED_INT },
	/* GPCOGL_CT_4_UINT32 */  { GL_RGBA32UI, GL_RGBA_INTEGER, GL_UNSIGNED_INT }
};

const char *vertex_shader =
	NL "#version 130"
	NL "in vec3 pos;"
	NL "void main()"
	NL "{"
	NL "	gl_Position = vec4(pos, 1);"
	NL "	gl_TexCoord[0].x = (gl_Position.x+1)/2;"
	NL "	gl_TexCoord[0].y = (gl_Position.y+1)/2;"
	NL "}";

static GLuint gpcogl_compile_shader(const char *shader_src, GLenum type)
{
	GLuint shid = glCreateShader(type);
	GLint comp_stat;

	if (shid == 0) {
		fprintf(stderr, "gpcogl_compile_shader(): error creating shaders!\n");
		return 0;
	}

	/* Attach source code to this object */
	glShaderSource(shid, 1, &shader_src, NULL);
	glCompileShader(shid);


	glGetShaderiv(shid, GL_COMPILE_STATUS, &comp_stat);
	if (!comp_stat) { /* failed */
		int length;
		glGetShaderiv(shid, GL_INFO_LOG_LENGTH, &length);
		char *msg = malloc(length+1);

		// Get additional information
		glGetShaderInfoLog(shid, length, &length, msg);
		fprintf(stderr, "gpcogl_compile_shader(): Failed to compile shader: %s\n", msg);
		glDeleteShader(shid);
		return 0;
	}

	return shid;
}

static GLuint gpcogl_link_program(GLuint vertex_shader_id, GLuint fragment_shader_id)
{
	GLuint progid = glCreateProgram();
	GLint lnink_stat;

	if (progid == 0) {
		fprintf(stderr, "gpcogl_link_program(): failed to create shader program\n");
		return 0;
	}

	glAttachShader(progid, vertex_shader_id);
	glAttachShader(progid, fragment_shader_id);

	glLinkProgram(progid);

	glGetProgramiv(progid, GL_LINK_STATUS, &lnink_stat);
	if (!lnink_stat) {
		fprintf(stderr, "gpcogl_link_program(): failed to link program\n");
		glDetachShader(progid, vertex_shader_id);
		glDetachShader(progid, fragment_shader_id);
		glDeleteProgram(progid);
		return 0;
	}

	return progid;
}

/* Load data in VBO (Vertex Buffer Object) and return its ID */
static GLuint gpcogl_load_vbo_rect(float sx, float sy)
{
	GLfloat vertices[6*3] = {0}; // triangle vertex coordinates
	GLuint vbo_id;

	/* rectangle made of two triangles */
	vertices[0*3 + 0] = -sx; vertices[0*3 + 1] = -sy;
	vertices[1*3 + 0] = -sx; vertices[1*3 + 1] = +sy;
	vertices[2*3 + 0] = +sx; vertices[2*3 + 1] = +sy;

	vertices[3*3 + 0] = -sx; vertices[3*3 + 1] = -sy;
	vertices[4*3 + 0] = +sx; vertices[4*3 + 1] = -sy;
	vertices[5*3 + 0] = +sx; vertices[5*3 + 1] = +sy;

	/* allocate buffer sapce and pass data to it */
	glGenBuffers(1, &vbo_id);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_id);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);

	/* unbind the active buffer */
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	return vbo_id;
}

GLuint gpcogl_redir_render_to_fb(gpcogl_t *ctx)
{
	GLuint fb;

	MAKE_CURRENT;

	glGenFramebuffers(1, &fb);                  /* create FBO (off-screen framebuffer) */
	glBindFramebuffer(GL_FRAMEBUFFER_EXT, fb);  /* bind offscreen buffer */

	return fb;
}


gpcogl_prog_id_t gpcogl_compile(gpcogl_t *ctx, const char *fragment_shader)
{
	GLuint vbo_id, vao_id;
	GLuint vshader_id, fshader_id, prog_id;
	GLuint posAttributePosition;

	MAKE_CURRENT;

	vbo_id = gpcogl_load_vbo_rect(1, 1);
	fshader_id = gpcogl_compile_shader(fragment_shader, GL_FRAGMENT_SHADER);
	if (fshader_id == 0)
		return 0;

	vshader_id = gpcogl_compile_shader(vertex_shader, GL_VERTEX_SHADER);
	prog_id = gpcogl_link_program(vshader_id, fshader_id);

	// Get the 'pos' variable location inside this program
	posAttributePosition = glGetAttribLocation(prog_id, "pos");

	/* Create and bind our vao it so that rest of vao operations affect this vao */
	glGenVertexArrays(1, &vao_id);
	glBindVertexArray(vao_id);

	/* buffer from which 'pos' will receive its data and the format of that data */
	glBindBuffer(GL_ARRAY_BUFFER, vbo_id);
	glVertexAttribPointer(posAttributePosition, 3, GL_FLOAT, false, 0, 0);

	/* Enable this attribute array linked to 'pos' */
	glEnableVertexAttribArray(posAttributePosition);

	return prog_id;
}

/* assumes MAKE_CURRENT */
static void compute(gpcogl_t *ctx, gpcogl_prog_id_t prog_id)
{
	static const int debug_half_only = 0;


	/* draw triangles starting from index 0 and using 3 or 6 indices for 1 or 2 triangles */
	glDrawArrays(GL_TRIANGLES, 0, debug_half_only ? 3 : 6);
}

void gpcogl_flush(gpcogl_t *ctx)
{
	/* swap the buffers and hence show the buffers content to the screen */
	if (ctx->flush == NULL)
		glFlush();
	else
		ctx->flush(ctx);
}

static int bind_texture(gpcogl_t *ctx, const char *id, int slot_no, GLuint texture_id, gpcogl_prog_id_t prg)
{
	static const int texture_target = GL_TEXTURE_RECTANGLE;

	glUseProgram(prg);
	GLuint uniaddr = glGetUniformLocation(prg, id);
	chkerr("glGetUniformLocation", -1);
	glUniform1i(uniaddr, slot_no);
	chkerr("glUniform1i", -1);


	/* tell texture unit GL_TEXTURE0 + slot_no to use texture0 */
	glActiveTexture(GL_TEXTURE0 + slot_no);
	glBindTexture(texture_target, texture_id);
	chkerr("glBindTexture1", -1);

	glTexParameteri(texture_target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(texture_target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(texture_target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(texture_target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	chkerr("glTexParameteri", -1);

	return 0;
}

gpcogl_array_id_t gpcogl_array_create(gpcogl_t *ctx, gpcogl_cell_type_t ct, int sx, int sy)
{
	const int texture_target = GL_TEXTURE_RECTANGLE;
	GLenum internal_format, texture_format, typename;
	int sketch = ctx->array_max_active-1;
	gpcogl_array_id_t tex_id;

	MAKE_CURRENT;

	if ((ct < 0) || (ct >= GPCOGL_CT_max))
		return gpcogl_no_array;

	internal_format = cell_spec[ct].intfmt;
	texture_format = cell_spec[ct].txtfmt;
	typename = cell_spec[ct].type;

	/* need to use a sketch pad for creating the texture */
	glActiveTexture(GL_TEXTURE0 + sketch);

	glGenTextures(1, &tex_id.tid);

	glBindTexture(texture_target, tex_id.tid);

	/* turn off filtering and set proper wrap mode */
	glTexParameteri(texture_target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(texture_target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(texture_target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(texture_target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	/* set texenv to replace instead of the default modulate */
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	/* and allocate graphics memory */
	glTexImage2D(texture_target, 0, internal_format, sx, sy, 0, texture_format, typename, 0);

	glBindTexture(texture_target, 0);

	tex_id.ct = ct;

	return tex_id;
}

static void reset_xform_to_texture(GLenum texture_target, GLint x0, GLint y0, GLsizei w, GLsizei h)
{
	static const int miplevel = 0;

	/* set up 1:1 transformation */
#if 0
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, sx, 0.0, sy);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
#endif

	if (w == 0) {
		GLint tw;
		glGetTexLevelParameteriv(texture_target, miplevel, GL_TEXTURE_WIDTH, &tw);
		w = tw - x0;
	}

	if (h == 0) {
		GLint th;
		glGetTexLevelParameteriv(texture_target, miplevel, GL_TEXTURE_HEIGHT, &th);
		h = th - y0;
	}

	/* printf("----VIEWPORT: %d %d\n", w, h); */
	glViewport(x0, y0, w, h);
}

/* Set compute output to the GPU array addresed by arr_id; arr_id is previosly
   created by gpcogl_array_create() */
static void gpcogl_set_output_array_box(gpcogl_t *ctx, gpcogl_array_id_t tex_id, GLint x0, GLint y0, GLsizei w, GLsizei h)
{
	const GLenum texture_target = GL_TEXTURE_RECTANGLE;

	MAKE_CURRENT;

	glActiveTexture(GL_TEXTURE0 + TEXTUNIT_STDOUT);

	glBindTexture(texture_target, tex_id.tid);

	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

	/* attach texture to the FBO */
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, texture_target, tex_id.tid, 0);

	reset_xform_to_texture(texture_target, x0, y0, w, h);
}

static void gpcogl_unset_output_array(gpcogl_t *ctx)
{
	const GLenum texture_target = GL_TEXTURE_RECTANGLE;

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, texture_target, 0, 0);
}


int gpcogl_array_get_size(gpcogl_t *ctx, GLint *sx, GLint *sy, gpcogl_array_id_t src)
{
	static const int texture_target = GL_TEXTURE_RECTANGLE;
	int sketch = ctx->array_max_active-1, level = 0;

	MAKE_CURRENT;

	glActiveTexture(GL_TEXTURE0 + sketch);
	chkerr("glActiveTexture", -1);

	glBindTexture(texture_target, src.tid);
	chkerr("glBindTexture", -1);

	glGetTexLevelParameteriv(texture_target, level, GL_TEXTURE_WIDTH, sx);
	glGetTexLevelParameteriv(texture_target, level, GL_TEXTURE_HEIGHT, sy);
	chkerr("glGetTexLevelParameteriv", -1);

	glBindTexture(texture_target, 0);

	if ((*sx == 0) || (*sy == 0))
		return -1;

	return 0;
}

int gpcogl_array_get(gpcogl_t *ctx, void *dst, gpcogl_array_id_t src)
{
	static const int texture_target = GL_TEXTURE_RECTANGLE;
	int sketch = ctx->array_max_active-1, level = 0;
	GLenum texture_format, typename;
	GLint sx, sy;

	if ((src.ct < 0) || (src.ct >= GPCOGL_CT_max) || (src.tid == -1))
		return -1;

	texture_format = cell_spec[src.ct].txtfmt;
	typename = cell_spec[src.ct].type;

	MAKE_CURRENT;

	glActiveTexture(GL_TEXTURE0 + sketch);
	chkerr("glActiveTexture", -1);

	glBindTexture(texture_target, src.tid);
	chkerr("glBindTexture", -1);

	/* it seems none of the above emits error for non-existent textures and
	   the only way to detect them is to check if they are 0-sized */
	glGetTexLevelParameteriv(texture_target, level, GL_TEXTURE_WIDTH, &sx);
	glGetTexLevelParameteriv(texture_target, level, GL_TEXTURE_HEIGHT, &sy);
	chkerr("glGetTexLevelParameteriv", -1);
	if ((sx == 0) || (sy == 0)) {
		glBindTexture(texture_target, 0);
		return -1;
	}

	glGetTexImage(texture_target, level, texture_format, typename, dst);
	chkerr("glGetTexImage", -1);


	glBindTexture(texture_target, 0);
	return 0;
}

int gpcogl_array_get_box(gpcogl_t *ctx, void *dst, gpcogl_array_id_t src, GLint x0, GLint y0, GLsizei w, GLsizei h)
{
	GLenum texture_format, typename;

	texture_format = cell_spec[src.ct].txtfmt;
	typename = cell_spec[src.ct].type;

	MAKE_CURRENT;
	
	/* can read pixels only from a frame buffer */
	gpcogl_set_output_array_box(ctx, src, 0, 0, 0, 0);
	glReadPixels(x0, y0, w, h, texture_format, typename, dst);
	gpcogl_unset_output_array(ctx);

	return 0;
}


gpcogl_array_id_t gpcogl_array_init(gpcogl_t *ctx, gpcogl_cell_type_t ct, const void *inp, int sx, int sy)
{
	gpcogl_array_id_t tex_id;
	static const int texture_target = GL_TEXTURE_RECTANGLE;
	GLenum internal_format, texture_format, typename;
	int sketch = ctx->array_max_active-1;

	if ((ct < 0) || (ct >= GPCOGL_CT_max))
		return gpcogl_no_array;

	internal_format = cell_spec[ct].intfmt;
	texture_format = cell_spec[ct].txtfmt;
	typename = cell_spec[ct].type;

	MAKE_CURRENT;

	glGenTextures(1, &tex_id.tid);
	chkerr("glGenTextures", gpcogl_no_array);

	/* +sketch is a sketch pad texture unit (used for the upload???) */
	glActiveTexture(GL_TEXTURE0 + sketch);
	glBindTexture(texture_target, tex_id.tid);
	glTexImage2D(texture_target, 0, internal_format, sx, sy, 0, texture_format, typename, inp);
	chkerr("glTexImage2D", gpcogl_no_array);
	glBindTexture(texture_target, 0);

	tex_id.ct = ct;

	return tex_id;
}


int gpcogl_init_context(gpcogl_t *ctx)
{
	MAKE_CURRENT;
	glewInit();

	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &ctx->array_max_size);
	glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS, &ctx->array_max_active);
	glGetIntegerv(GL_MAJOR_VERSION, &ctx->glver_major);
	glGetIntegerv(GL_MINOR_VERSION, &ctx->glver_minor);
	ctx->gl_vendor = glGetString(GL_VENDOR);
	ctx->renderer_vendor = glGetString(GL_RENDERER);

	if (ctx->glver_major < 3)
		return -1;
	return 0;
}

void gpcogl_uninit_context(gpcogl_t *ctx)
{
	ctx->array_max_size = 0;
	ctx->array_max_active = 0;
}


static void unbind_all(gpcogl_t *ctx,int limit)
{
	const int texture_target = GL_TEXTURE_RECTANGLE;
	int n, sketch = ctx->array_max_active-1;

	for(n = 0; n < limit; n++) {
		glActiveTexture(GL_TEXTURE0 + n);
		glBindTexture(texture_target, 0);
	}
	glActiveTexture(GL_TEXTURE0 + sketch);

}

const char GPCOGL_C_1F[] = "const 1*float";
const char GPCOGL_C_2F[] = "const 2*float";
const char GPCOGL_C_3F[] = "const 3*float";
const char GPCOGL_C_4F[] = "const 4*float";
const char GPCOGL_C_1U[] = "const 1*uint32";
const char GPCOGL_C_2U[] = "const 2*uint32";
const char GPCOGL_C_3U[] = "const 3*uint32";
const char GPCOGL_C_4U[] = "const 4*uint32";
const char GPCOGL_C_1I[] = "const 1*int32";
const char GPCOGL_C_2I[] = "const 2*int32";
const char GPCOGL_C_3I[] = "const 3*int32";
const char GPCOGL_C_4I[] = "const 4*int32";

#define setup_uniform(name, glcall, len, data) \
do { \
	GLuint uniaddr = glGetUniformLocation(pid, name); \
	glcall; \
} while(0)

#define setup_uniform_a(name, glcall, len, typ, arr) \
do { \
	const typ *data = arr; \
	setup_uniform(name, glcall, len, data); \
} while(0)

int gpcogl_acompute_box(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, GLint x0, GLint y0, GLsizei w, GLsizei h, const char **src_name, gpcogl_array_id_t *src_aid, const char **const_type, const char **const_name, const void **const_val)
{
	int tu = 1, res = 0, sketch = ctx->array_max_active-1;;

	MAKE_CURRENT;

	if (dst.tid != -1)
		gpcogl_set_output_array_box(ctx, dst, x0, y0, w, h);

	glUseProgram(pid);

	if (src_name != NULL) {
		for(tu = 1;;tu++) {
			const char *name = *src_name;
			gpcogl_array_id_t aid = *src_aid;

			src_name++;
			src_aid++;

			if (name == NULL)
				break;

			if (tu >= sketch) {
				/* TODO: error message for too many input arrays */
				res = -1;
				goto error;
			}

			bind_texture(ctx, name, tu, aid.tid, pid);
		}
	}

	if (const_type != NULL) {
		int n;
		for(n = 0; const_type[n] != NULL; n++) {
			const char *typ = const_type[n], *nam = const_name[n];
			const void *dat = const_val[n];

			if (typ == GPCOGL_C_1F)      setup_uniform_a(nam, glUniform1f(uniaddr, data[0]), 1, float, dat);
			else if (typ == GPCOGL_C_2F) setup_uniform_a(nam, glUniform2f(uniaddr, data[0], data[1]), 2, float, dat);
			else if (typ == GPCOGL_C_3F) setup_uniform_a(nam, glUniform3f(uniaddr, data[0], data[1], data[2]), 3, float, dat);
			else if (typ == GPCOGL_C_4F) setup_uniform_a(nam, glUniform4f(uniaddr, data[0], data[1], data[2], data[3]), 4, float, dat);
			else if (typ == GPCOGL_C_1U) setup_uniform_a(nam, glUniform1ui(uniaddr, data[0]), 1, uint32_t, dat);
			else if (typ == GPCOGL_C_2U) setup_uniform_a(nam, glUniform2ui(uniaddr, data[0], data[1]), 2, uint32_t, dat);
			else if (typ == GPCOGL_C_3U) setup_uniform_a(nam, glUniform3ui(uniaddr, data[0], data[1], data[2]), 3, uint32_t, dat);
			else if (typ == GPCOGL_C_4U) setup_uniform_a(nam, glUniform4ui(uniaddr, data[0], data[1], data[2], data[3]), 4, uint32_t, dat);
			else if (typ == GPCOGL_C_1I) setup_uniform_a(nam, glUniform1i(uniaddr, data[0]), 1, int32_t, dat);
			else if (typ == GPCOGL_C_2I) setup_uniform_a(nam, glUniform2i(uniaddr, data[0], data[1]), 2, int32_t, dat);
			else if (typ == GPCOGL_C_3I) setup_uniform_a(nam, glUniform3i(uniaddr, data[0], data[1], data[2]), 3, int32_t, dat);
			else if (typ == GPCOGL_C_4I) setup_uniform_a(nam, glUniform4i(uniaddr, data[0], data[1], data[2], data[3]), 4, int32_t, dat);
			else
				fprintf(stderr, "gpcogl_acompute_box(): unknown type %s\n", typ);
		}
	}

	compute(ctx, pid);
	if (dst.tid != -1)
		gpcogl_unset_output_array(ctx); /* "flush" output - finish texture before anything else reads it */

	error:;
	unbind_all(ctx, tu);
	return res;
}

int gpcogl_acompute(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, const char **src_name, gpcogl_array_id_t *src_aid, const char **const_type, const char **const_name, const void **const_val)
{
	return gpcogl_acompute_box(ctx, pid, dst, 0, 0, 0, 0, src_name, src_aid, const_type, const_name, const_val);
}


#define setup_uniform_v(ap, glcall, len, typ) \
do { \
	const char *name = va_arg(ap, const char *); \
	const typ *data = va_arg(ap, const typ *); \
	setup_uniform(name, glcall, len, data); \
} while(0)


static int vcompute_box(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, GLint x0, GLint y0, GLsizei w, GLsizei h, va_list ap)
{
	int tu, res = 0, sketch = ctx->array_max_active-1;

	MAKE_CURRENT;

	if (dst.tid != -1)
		gpcogl_set_output_array_box(ctx, dst, x0, y0, w, h);

	glUseProgram(pid);

	for(tu = 1;;tu++) {
		const char *name;
		gpcogl_array_id_t aid;

		name = va_arg(ap, const char *);
		if (name == NULL)
			break;

		if (name == GPCOGL_C_1F)      setup_uniform_v(ap, glUniform1f(uniaddr, data[0]), 1, float);
		else if (name == GPCOGL_C_2F) setup_uniform_v(ap, glUniform2f(uniaddr, data[0], data[1]), 2, float);
		else if (name == GPCOGL_C_3F) setup_uniform_v(ap, glUniform3f(uniaddr, data[0], data[1], data[2]), 3, float);
		else if (name == GPCOGL_C_4F) setup_uniform_v(ap, glUniform4f(uniaddr, data[0], data[1], data[2], data[3]), 4, float);
		else if (name == GPCOGL_C_1U) setup_uniform_v(ap, glUniform1ui(uniaddr, data[0]), 1, uint32_t);
		else if (name == GPCOGL_C_2U) setup_uniform_v(ap, glUniform2ui(uniaddr, data[0], data[1]), 2, uint32_t);
		else if (name == GPCOGL_C_3U) setup_uniform_v(ap, glUniform3ui(uniaddr, data[0], data[1], data[2]), 3, uint32_t);
		else if (name == GPCOGL_C_4U) setup_uniform_v(ap, glUniform4ui(uniaddr, data[0], data[1], data[2], data[3]), 4, uint32_t);
		else if (name == GPCOGL_C_1I) setup_uniform_v(ap, glUniform1i(uniaddr, data[0]), 1, int32_t);
		else if (name == GPCOGL_C_2I) setup_uniform_v(ap, glUniform2i(uniaddr, data[0], data[1]), 2, int32_t);
		else if (name == GPCOGL_C_3I) setup_uniform_v(ap, glUniform3i(uniaddr, data[0], data[1], data[2]), 3, int32_t);
		else if (name == GPCOGL_C_4I) setup_uniform_v(ap, glUniform4i(uniaddr, data[0], data[1], data[2], data[3]), 4, int32_t);
		else {
			/* input array */
			aid = va_arg(ap, gpcogl_array_id_t);

			if (tu >= sketch) {
				/* TODO: error message for too many input arrays */
				res = -1;
				goto error;
			}

			bind_texture(ctx, name, tu, aid.tid, pid);
		}
	}

	compute(ctx, pid);
	if (dst.tid != -1)
		gpcogl_unset_output_array(ctx); /* "flush" output - finish texture before anything else reads it */

	error:;
	unbind_all(ctx, tu);
	return res;
}

int gpcogl_vcompute(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, ...)
{
	int res;
	va_list ap;

	va_start(ap, dst);
	res = vcompute_box(ctx, pid, dst, 0, 0, 0, 0, ap);
	va_end(ap);

	return res;
}

int gpcogl_vcompute_box(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, GLint x0, GLint y0, GLsizei w, GLsizei h, ...)
{
	int res;
	va_list ap;

	va_start(ap, h);
	res = vcompute_box(ctx, pid, dst, x0, y0, w, h, ap);
	va_end(ap);

	return res;
}


int gpcogl_array_clear(gpcogl_t *ctx, gpcogl_array_id_t dst)
{
	MAKE_CURRENT;

	/* bind dst to the output fb */
	gpcogl_set_output_array_box(ctx, dst, 0, 0, 0, 0);

	/* clear the output fb */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT);

	gpcogl_unset_output_array(ctx); /* "flush" output - finish texture before anything else reads it */

	return 0;
}

int gpcogl_array_free(gpcogl_t *ctx, gpcogl_array_id_t aid)
{
	MAKE_CURRENT;

	glDeleteTextures(1, &aid.tid);
	chkerr("glDeleteTextures", -1);

	return 0;
}

void gpcogl_finish(gpcogl_t *ctx)
{
	MAKE_CURRENT;
	glFinish();
}

